#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkJunctionDetectionFilter.h"

class CommandProgressUpdate : public itk::Command 
{
public:
  typedef  CommandProgressUpdate    Self;
  typedef  itk::Command             Superclass;
  typedef  itk::SmartPointer<CommandProgressUpdate>  Pointer;
  itkNewMacro( CommandProgressUpdate );

  void Execute(itk::Object *caller, const itk::EventObject & event)
  {
    Execute( (const itk::Object *)caller, event);
  }

  void Execute(const itk::Object * object, const itk::EventObject & event)
  {
    if( !(itk::ProgressEvent().CheckEvent( &event )) )  return;
    std::cout << "=";
  }

protected:
  CommandProgressUpdate(){};
};

int main( int argc, char* argv[] )
{
  if( argc < 5 )
  {
    std::cerr << "Usage: " << std::endl;
    std::cerr << argv[0] << " inputImage outputImage inner outer" << std::endl;
    std::cerr << "Example: " << std::endl;
    std::cerr << argv[0] << " input.mha output.mha 2.0 3.0" << std::endl;
    return EXIT_FAILURE;
  }
  
  typedef itk::Image<unsigned short, 2>                       ImageType;
  typedef itk::ImageFileReader<ImageType>                     ReaderType;
  typedef itk::JunctionDetectionFilter<ImageType>             DetectorType;
  typedef itk::ImageFileWriter<DetectorType::OutputImageType> WriterType;

  char *inputfilename = argv[1];
  char *outputfilename = argv[2];
  float inner = atof(argv[3]);
  float outer = atof(argv[4]);

  ImageType::Pointer inputImage;
  ReaderType::Pointer reader = ReaderType::New();
  reader->SetFileName( inputfilename );
  try 
  { 
    reader->Update();
    inputImage = reader->GetOutput();
  } 
  catch( itk::ExceptionObject & err ) 
  { 
    std::cerr << "ExceptionObject caught !" << std::endl; 
    std::cerr << err << std::endl; 
    return EXIT_FAILURE;
  }

  // Junction detection with a progress observer
  CommandProgressUpdate::Pointer observer = CommandProgressUpdate::New();
  DetectorType::Pointer detector = DetectorType::New();
  detector->SetInnerRadius(inner);
  detector->SetOuterRadius(outer);
  detector->SetMinNumberOfPixel(6);
  detector->SetInput(reader->GetOutput());
  detector->AddObserver(itk::ProgressEvent(), observer);
  detector->Update();

  WriterType::Pointer writer = WriterType::New();
  writer->SetFileName( outputfilename );
  writer->SetInput( detector->GetOutput() );
  try 
  { 
    writer->Update(); 
  } 
  catch( itk::ExceptionObject & err ) 
  { 
    std::cerr << "ExceptionObject caught !" << std::endl; 
    std::cerr << err << std::endl; 
    return EXIT_FAILURE;
  }

  std::cout << std::endl;
  DetectorType::JCLabelMapType& jcLabelmap = detector->GetJCLabelMap();
  for( DetectorType::JCLabelMapType::iterator jclmIt = jcLabelmap.begin(); jclmIt!=jcLabelmap.end(); ++jclmIt )
  {
    int jcLabel = (*jclmIt).first;
    DetectorType::JCLabelPairType jclPair = (*jclmIt).second;
    ImageType::IndexType jcIndex = jclPair.first;
    ImageType::PointType point;
    inputImage->TransformIndexToPhysicalPoint(jcIndex, point);
    float radius = jclPair.second;
    std::cout << "Junction " << jcLabel << " at " << point << " size " << radius << std::endl;
  }

  return EXIT_SUCCESS;
}
