#include "itkImageFileReader.h"
#include "itkImageFileWriter.h"
#include "itkJunctionDetectionFilter.h"

#include "vtkPolyData.h"
#include "vtkXMLPolyDataWriter.h"
#include "vtkSphereSource.h"
#include "vtkAppendPolyData.h"

class CommandProgressUpdate : public itk::Command 
{
public:
  typedef  CommandProgressUpdate    Self;
  typedef  itk::Command             Superclass;
  typedef  itk::SmartPointer<CommandProgressUpdate>  Pointer;
  itkNewMacro( CommandProgressUpdate );

  void Execute(itk::Object *caller, const itk::EventObject & event)
  {
    Execute( (const itk::Object *)caller, event);
  }

  void Execute(const itk::Object * object, const itk::EventObject & event)
  {
    if( !(itk::ProgressEvent().CheckEvent( &event )) )  return;
    std::cout << "=";
  }

protected:
  CommandProgressUpdate(){};
};

int main( int argc, char* argv[] )
{
  if( argc < 6 )
  {
    std::cerr << "Usage: " << std::endl;
    std::cerr << argv[0] << " inputImage outputImage junctionspheres inner outer" << std::endl;
    std::cerr << "Example: " << std::endl;
    std::cerr << argv[0] << " input.mha output.mha jcspheres.vtp 2.0 3.0" << std::endl;
    return EXIT_FAILURE;
  }
  
  typedef itk::Image<unsigned short, 3>                       ImageType;
  typedef itk::ImageFileReader<ImageType>                     ReaderType;
  typedef itk::JunctionDetectionFilter<ImageType>             DetectorType;
  typedef itk::ImageFileWriter<DetectorType::OutputImageType> WriterType;

  char *inputfilename = argv[1];
  char *outputfilename = argv[2];
  char *jcspheresfilename = argv[3];
  float inner = atof(argv[4]);
  float outer = atof(argv[5]);

  ImageType::Pointer inputImage;
  ReaderType::Pointer reader = ReaderType::New();
  reader->SetFileName( inputfilename );
  try 
  { 
    reader->Update();
    inputImage = reader->GetOutput();
  } 
  catch( itk::ExceptionObject & err ) 
  { 
    std::cerr << "ExceptionObject caught !" << std::endl; 
    std::cerr << err << std::endl; 
    return EXIT_FAILURE;
  }

  // Junction detection with a progress observer
  CommandProgressUpdate::Pointer observer = CommandProgressUpdate::New();
  DetectorType::Pointer detector = DetectorType::New();
  detector->SetInnerRadius(inner);
  detector->SetOuterRadius(outer);
  detector->SetMinNumberOfPixel(16);
  detector->SetInput(reader->GetOutput());
  detector->AddObserver(itk::ProgressEvent(), observer);
  detector->Update();

  WriterType::Pointer writer = WriterType::New();
  writer->SetFileName( outputfilename );
  writer->SetInput( detector->GetOutput() );
  try 
  { 
    writer->Update(); 
  } 
  catch( itk::ExceptionObject & err ) 
  { 
    std::cerr << "ExceptionObject caught !" << std::endl; 
    std::cerr << err << std::endl; 
    return EXIT_FAILURE;
  }
  
  // Write junctions as spheres
  vtkPolyData *vPoly = vtkPolyData::New();
  vtkSphereSource *vSphere = vtkSphereSource::New();
  vSphere->SetPhiResolution(32);
  vSphere->SetThetaResolution(32);
  vtkAppendPolyData *vAppend = vtkAppendPolyData::New();
  std::cout << std::endl;
  DetectorType::JCLabelMapType& jcLabelmap = detector->GetJCLabelMap();
  for( DetectorType::JCLabelMapType::iterator jclmIt = jcLabelmap.begin(); jclmIt!=jcLabelmap.end(); ++jclmIt )
  {
    int jcLabel = (*jclmIt).first;
    DetectorType::JCLabelPairType jclPair = (*jclmIt).second;
    ImageType::IndexType jcIndex = jclPair.first;
    ImageType::PointType point;
    inputImage->TransformIndexToPhysicalPoint(jcIndex, point);
    float radius = jclPair.second;
    std::cout << "Junction " << jcLabel << " at " << point << " size " << radius << std::endl;
    vSphere->SetCenter(point[0], point[1], point[2]);
    vSphere->SetRadius(radius);
    vSphere->Update();
    vAppend->AddInput(vPoly);
    vAppend->AddInput(vSphere->GetOutput());
    vAppend->Update();
    vAppend->RemoveInput(vPoly);
    vAppend->RemoveInput(vSphere->GetOutput());
    vPoly->Reset();
    vPoly->Squeeze();
    vPoly->DeepCopy(vAppend->GetOutput());
  }
  vSphere->Delete();
  vAppend->Delete();

  // Save spheres into a vtp file
  vtkXMLPolyDataWriter *vPolywriter = vtkXMLPolyDataWriter::New();
  vPolywriter->SetInput(vPoly);
  vPolywriter->SetFileName(jcspheresfilename);
  vPolywriter->SetDataModeToBinary();
  vPolywriter->Write();
  vPolywriter->Delete();
  vPoly->Delete();

  return EXIT_SUCCESS;
}
